import requests

from graphw00f.helpers import error_contains

requests.packages.urllib3.disable_warnings()

class GraphQLDetectionFailed(Exception):
  pass

class GraphQLError(Exception):
  pass

class GRAPHW00F:
  def __init__(self, headers,
                     cookies,
                     timeout,
                     proxies,
                     follow_redirects=False):
    self.url = 'http://example.com'
    self.cookies = cookies
    self.headers = headers
    self.follow_redirects = follow_redirects,
    self.timeout = timeout
    self.proxies = proxies

  def check(self, url):
    query = '''
      query {
        __typename
      }
    '''
    response = self.graph_query(url, payload=query)
    if response.get('data'):
      if response.get('data', {}).get('__typename', '') in ('Query', 'QueryRoot', 'query_root'):
        return True
    elif response.get('errors') and (any('locations' in i for i in response['errors']) or any('extensions' in i for i in response['errors'])):
      return True
    elif response.get('data'):
      return True
    else:
      raise GraphQLDetectionFailed

  def execute(self, url):
    self.url = url
    if self.engine_inigo():
      return 'inigo'
    elif self.engine_lighthouse():
      return 'lighthouse'
    elif self.engine_caliban():
      return 'caliban'
    elif self.engine_lacinia():
      return 'lacinia'
    elif self.engine_jaal():
      return 'jaal'
    elif self.engine_morpheus():
      return 'morpheus-graphql'
    elif self.engine_mercurius():
      return 'mercurius'
    elif self.engine_graphql_yoga():
      return 'graphql_yoga'
    elif self.engine_agoo():
      return 'agoo'
    elif self.engine_tailcall():
      return 'tailcall'
    elif self.engine_dgraph():
      return 'dgraph'
    elif self.engine_graphene():
      return 'graphene'
    elif self.engine_ariadne():
      return 'ariadne'
    elif self.engine_apollo():
      return 'apollo'
    elif self.engine_awsappsync():
      return 'aws-appsync'
    elif self.engine_hasura():
      return 'hasura'
    elif self.engine_wpgraphql():
      return 'wpgraphql'
    elif self.engine_graphqlapiforwp():
      return 'graphql-api-for-wp'
    elif self.engine_graphqljava():
      return 'graphql-java'
    elif self.engine_hypergraphql():
      return 'hypergraphql'
    elif self.engine_ruby():
      return 'ruby-graphql'
    elif self.engine_graphqlphp():
      return 'graphql-php'
    elif self.engine_gqlgen():
      return 'gqlgen'
    elif self.engine_graphqlgo():
      return 'graphql-go'
    elif self.engine_juniper():
      return 'juniper'
    elif self.engine_sangria():
      return 'sangria'
    elif self.engine_dianajl():
      return 'dianajl'
    elif self.engine_strawberry():
      return 'strawberry'
    elif self.engine_tartiflette():
      return 'tartiflette'
    elif self.engine_directus():
      return 'directus'
    elif self.engine_absinthe():
      return 'absinthe-graphql'
    elif self.engine_graphqldotnet():
      return 'graphql-dotnet'
    elif self.engine_pggraphql():
      return 'pg_graphql'
    elif self.engine_hotchocolate():
      return 'hotchocolate'
    elif self.engine_ballerina():
      return 'ballerina'
    elif self.engine_flutter():
      return 'flutter'
    
    return None

  def graph_query(self, url, operation='query', payload={}):
    try:
      # GraphQL may support POST or GET, or both. Choose the one that returns a JSON.
      response_post = requests.post(url,
                             headers=self.headers,
                             cookies=self.cookies,
                             verify=False,
                             allow_redirects=self.follow_redirects,
                             timeout=self.timeout,
                             proxies=self.proxies,
                             json={operation:payload})
      response_get = requests.get(url,
                             headers=self.headers,
                             cookies=self.cookies,
                             verify=False,
                             allow_redirects=self.follow_redirects,
                             timeout=self.timeout,
                             proxies=self.proxies,
                             params={operation:payload})
      if isinstance(response_post.json(), dict):
        return response_post.json()
      elif isinstance(response_get.json(), dict):
        return response_get.json()
      else:
        return {}
    except:
      return {}

  def engine_graphql_yoga(self):
    query = '''
      subscription {
         __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'asyncExecutionResult[Symbol.asyncIterator] is not a function') or error_contains(response, 'Unexpected error.'):
        return True

    return False
  def engine_apollo(self):
    query = '''
      query @skip {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive "@skip" argument "if" of type "Boolean!" is required, but it was not provided.'):
      return True

    query = '''
      query @deprecated {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive "@deprecated" may not be used on QUERY.'):
      return True

    return False

  def engine_awsappsync(self):
      query = 'query @skip { __typename }'
      response = self.graph_query(self.url, payload=query)
      return error_contains(response, 'MisplacedDirective')

  def engine_graphene(self):
    query = '''aaa'''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Syntax Error GraphQL (1:1)'):
      return True

    return False

  def engine_hasura(self):
    query = '''
      query @cached {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if response.get('data'):
      if response.get('data', {}).get('__typename', '') == 'query_root':
        return True
    query = '''
     query {
       aaa
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'field "aaa" not found in type: \'query_root\''):
      return True

    query = '''
      query @skip {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'directive "skip" is not allowed on a query'):
      return True

    query = '''
      query {
        __schema
      }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'missing selection set for "__Schema"'):
      return True

    return False

  def engine_graphqlphp(self):
    query = '''
      query ! {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Syntax Error: Cannot parse the unexpected character "?".'):
      return True

    query = '''
      query @deprecated {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive "deprecated" may not be used on "QUERY".'):
      return True

    return False

  def engine_ruby(self):
    query = '''
     query @skip {
       __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, '\'@skip\' can\'t be applied to queries (allowed: fields, fragment spreads, inline fragments)'):
      return True
    elif error_contains(response, 'Directive \'skip\' is missing required arguments: if'):
      return True

    query = '''
     query @deprecated {
       __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, '\'@deprecated\' can\'t be applied to queries'):
      return True

    query = '''
      query {
       __typename {
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Parse error on "}" (RCURLY)'):
      return True

    query = '''
      query {
        __typename @skip
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive \'skip\' is missing required arguments: if'):
      return True

    return False

  def engine_hypergraphql(self):
    query = '''
     zzz {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Validation error of type InvalidSyntax: Invalid query syntax.'):
      return True

    query = '''
      query {
        alias1:__typename @deprecated
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Validation error of type UnknownDirective: Unknown directive deprecated @ \'__typename\''):
      return True

    return False

  def engine_graphqljava(self):
    query = '''
     queryy  {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Invalid Syntax : offending token \'queryy\''):
      return True

    query = '''
     query @aaa@aaa {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Validation error of type DuplicateDirectiveName: Directives must be uniquely named within a location.'):
      return True

    query = ''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Invalid Syntax : offending token \'<EOF>\''):
      return True

    return False

  def engine_ariadne(self):
    query = '''
      query {
        __typename @abc
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Unknown directive \'@abc\'.') and 'data' not in response:
      return True

    query = ''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'The query must be a string.'):
      return True

    return False

  def engine_graphqlapiforwp(self):
    query = '''
     query {
       alias1$1:__typename
     }
    '''
    response = self.graph_query(self.url, payload=query)
    if response.get('data'):
      if response.get('data', {}).get('alias1$1', '') == 'QueryRoot':
        return True

    query = '''query aa#aa { __typename }'''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Unexpected token "END"'):
      return True

    query = '''
      query @skip {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Argument \'if\' cannot be empty, so directive \'skip\' has been ignored'):
      return True

    query = '''
      query @doesnotexist {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'No DirectiveResolver resolves directive with name \'doesnotexist\''):
      return True

    query = ''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'The query in the body is empty'):
      return True

    return False

  def engine_wpgraphql(self):
    query = ''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'GraphQL Request must include at least one of those two parameters: "query" or "queryId"'):
      return True

    query = '''
     query {
       alias1$1:__typename
     }
    '''
    response = self.graph_query(self.url, payload=query)
    if not error_contains(response, 'Syntax Error: Expected Name, found $'):
      return False

    try:
      debug_msg = response['extensions']['debug'][0]
      if debug_msg['type'] == 'DEBUG_LOGS_INACTIVE' or \
        debug_msg['message'] == 'GraphQL Debug logging is not active. To see debug logs, GRAPHQL_DEBUG must be enabled.':
        return True
    except KeyError:
      pass

    return False

  def engine_gqlgen(self):
    query = '''
      query  {
      __typename {
    }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'expected at least one definition'):
      return True

    query = '''
      query  {
      alias^_:__typename {
    }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Expected Name, found <Invalid>'):
      return True

    return False

  def engine_graphqlgo(self):
    query = '''
      query {
      __typename {
      }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Unexpected empty IN'):
      return True

    query = ''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Must provide an operation.'):
      return True

    query = '''
      query {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    try:
      if response['data'] != None and response['data']['__typename'] == 'RootQuery':
        return True
    except KeyError:
      pass

    return False

  def engine_juniper(self):
    query = '''
      queryy {
        __typename
    }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Unexpected "queryy"'):
      return True

    query = ''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Unexpected end of input'):
      return True

    return False

  def engine_sangria(self):
    query = '''
      queryy {
        __typename
    }
    '''
    response = self.graph_query(self.url, payload=query)
    syntaxError = response.get('syntaxError', '')
    if 'Syntax error while parsing GraphQL query. Invalid input "queryy", expected ExecutableDefinition or TypeSystemDefinition' in syntaxError:
      return True

    return False

  def engine_flutter(self):
    query = '''
      query {
        __typename @deprecated
    }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive "deprecated" may not be used on FIELD.'):
      return True

    return False

  def engine_dianajl(self):
    query = '''queryy { __typename }'''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Syntax Error GraphQL request (1:1) Unexpected Name "queryy"'):
      return True

    return False

  def engine_strawberry(self):
    query = '''
      query @deprecated {
        __typename
      }'''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive \'@deprecated\' may not be used on query.') and 'data' in response:
      return True

    return False

  def engine_tartiflette(self):
    query = '''
      query @a { __typename }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Unknow Directive < @a >.'):
      return True

    query = '''
      query @skip { __typename }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Missing mandatory argument < if > in directive < @skip >.'):
      return True

    query = '''
      query { graphwoof }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Field graphwoof doesn\'t exist on Query'):
      return True

    query = '''
      query {
        __typename @deprecated
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Directive < @deprecated > is not used in a valid location.'):
      return True

    query = '''
      queryy {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'syntax error, unexpected IDENTIFIER'):
      return True

    return False

  def engine_tailcall(self):
    query = '''
      aa {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'expected executable_definition'):
      return True

    return False


  def engine_dgraph(self):
    query = '''
      query {
        __typename @cascade
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if 'data' in response and response['data']:
      if response.get('data').get('__typename', '') == 'Query':
        return True

    query = '''
      query {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Not resolving __typename. There\'s no GraphQL schema in Dgraph. Use the /admin API to add a GraphQL schema'):
      return True

    return False

  def engine_directus(self):
    query = ''

    response = self.graph_query(self.url, payload=query)
    errors = response.get('errors', [])
    if response.get('errors', []):
      if errors and errors[0].get('extensions', {}).get('code' '') == 'INVALID_PAYLOAD':
        return True

    return False

  def engine_lighthouse(self):
    query = '''
      query {
        __typename @include(if: falsee)
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Internal server error') or error_contains(response, 'internal', part='category'):
      return True

    return False

  def engine_agoo(self):
    query = '''
      query {
        zzz
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'eval error', part='code'):
      return True

    return False

  def engine_mercurius(self):
    query = ''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Unknown query'):
      return True
    return False

  def engine_morpheus(self):
    query = '''
      queryy {
          __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'expecting white space') or error_contains(response, 'offset'):
      return True

    return False

  def engine_lacinia(self):
    query = '''
      query {
        graphw00f
      }
    '''

    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Cannot query field `graphw00f\' on type `QueryRoot\'.'):
        return True

    return False

  def engine_jaal(self):
    query = '''{}'''
    response = self.graph_query(self.url, payload=query, operation='{}')

    if error_contains(response, 'must have a single query') or error_contains(response, 'offset'):
      return True

    return False

  def engine_caliban(self):
    query = '''
      query {
        __typename
      }

      fragment woof on __Schema {
        directives {
          name
        }
      }
    '''

    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Fragment \'woof\' is not used in any spread'):
      return True

    return False

  def engine_absinthe(self):
    query = '''
      query {
        graphw00f
      }
    '''

    response = self.graph_query(self.url, payload=query)

    if error_contains(response, 'Cannot query field \"graphw00f\" on type \"RootQueryType\".'):
        return True

    return False

  def engine_graphqldotnet(self):
    query = 'query @skip { __typename }'
    response = self.graph_query(self.url, payload=query)
    return error_contains(response, 'Directive \'skip\' may not be used on Query.')
  
  def engine_pggraphql(self):
    query = '''query { __typename @skip(aa:true) }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Unknown argument to @skip: aa'):
      return True
    
    return False

  def engine_hotchocolate(self):
    query = '''
    queryy  {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Unexpected token: Name.'):
      return True

    query = '''
    query @aaa@aaa {
        __typename
      }
    '''
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'The specified directive `aaa` is not supported by the current schema.'):
      return True

    return False
  
  def engine_inigo(self):
      query = '''
      query  {
          __typename
        }
      '''
      response = self.graph_query(self.url, payload=query)
      if 'extensions' in response and 'inigo' in response['extensions']:
          return True

      return False
  
  def engine_ballerina(self):
    query = '''
    query {
      __typename
      ...A
    }

    fragment A on Query {
      ...B
    }

    fragment B on Query {
      ...A
    }
    '''
    
    response = self.graph_query(self.url, payload=query)
    if error_contains(response, 'Cannot spread fragment "A" within itself via "B"'):
        return True

    return False
  



import datetime
import os.path
from urllib.parse import urlparse
from version import VERSION

class bcolors:
  OKBLUE = '\033[94m'
  OKCYAN = '\033[96m'
  OKGREEN = '\033[92m'
  WARNING = '\033[93m'
  FAIL = '\033[91m'
  ENDC = '\033[0m'

def read_custom_wordlist(location):
  wordlists = set()
  if os.path.exists(location):
    f = open(location, 'r').read()
    for line in f.splitlines():
      if not line.startswith('/'):
        line = '/' + line

      wordlists.add(line)
  else:
    print('Could not find wordlist file: {}'.format(location))
  return wordlists

def error_contains(response, word_to_match, part='message'):
    word_to_match = word_to_match.lower() 
    
    if isinstance(response, dict):
        errors = response.get('errors', [])
        if isinstance(errors, list):
            return any(
                word_to_match in (error.get(part, '') if isinstance(error, dict) else str(error)).lower()
                for error in errors
            )
        elif isinstance(errors, str):
            return word_to_match in errors.lower()
    elif isinstance(response, str):
        return word_to_match in response.lower()
    return False

def get_time():
  return datetime.datetime.now().strftime('%Y-%m-%d')

def draw_art():
  return '''
                +-------------------+
                |     graphw00f     |
                +-------------------+
                  ***            ***
                **                  **
              **                      **
    +--------------+              +--------------+
    |    Node X    |              |    Node Y    |
    +--------------+              +--------------+
                  ***            ***
                     **        **
                       **    **
                    +------------+
                    |   Node Z   |
                    +------------+

                graphw00f - v{version}
          The fingerprinting tool for GraphQL
           Dolev Farhi <dolev@lethalbit.com>
  '''.format(version=VERSION)

def possible_graphql_paths():
  return [
    '',
    '/',
    '/api',
    '/graphql',
    '/graphiql',
    '/v1/graphql',
    '/v2/graphql',
    '/v3/graphql',
    '/v1/graphiql',
    '/v2/graphiql',
    '/v3/graphiql',
    '/graphql/v1',
    '/graphql/v2',
    '/graphql/v3',
    '/api/graphql',
    '/api/graphiql',
    '/console',
    '/playground',
    '/gql',
    '/query',
    '/index.php?graphql',
    '/rpc/graphql'
  ]

def get_engines():
  return {
    'apollo':{
      'name':'Apollo',
      'url':'https://www.apollographql.com',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/apollo.md',
      'technology':['JavaScript', 'Node.js', 'TypeScript']
    },
    'aws-appsync':{
      'name':'AWS AppSync',
      'url':'https://aws.amazon.com/appsync',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/appsync.md',
      'technology':[],
    },
    'graphene':{
      'name':'Graphene',
      'url':'https://graphene-python.org',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphene.md',
      'technology':['Python']
    },
    'hasura':{
      'name':'Hasura',
      'url':'https://hasura.io',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/hasura.md',
      'technology':['Haskell']
    },
    'graphql-php':{
      'name':'GraphQL PHP',
      'url':'https://webonyx.github.io/graphql-php',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-php.md',
      'technology':['PHP']
    },
    'ruby-graphql':{
      'name':'Ruby GraphQL',
      'url':'https://graphql-ruby.org',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-ruby.md',
      'technology':['Ruby']
    },
    'hypergraphql':{
      'name':'HyperGraphQL',
      'url':'https://www.hypergraphql.org',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/hypergraphql.md',
      'technology':['Java']
    },
    'ariadne':{
      'name':'Ariadne',
      'url':'https://ariadnegraphql.org',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/ariadne.md',
      'technology':['Python']
    },
    'graphql-api-for-wp':{
      'name':'GraphQL API for Wordpress (Gato GraphQL)',
      'url':'https://graphql-api.com',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-api-for-wp.md',
      'technology':['PHP'],
    },
    'wpgraphql':{
      'name':'WPGraphQL WordPress Plugin',
      'url':'https://www.wpgraphql.com',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/wp-graphql.md',
      'technology':['PHP']
    },
    'gqlgen':{
      'name':'gqlgen - GraphQL for Go',
      'url':'https://gqlgen.com',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/gqlgen.md',
      'technology':['Go']
    },
    'graphql-go':{
      'name':'graphql-go -GraphQL for Go',
      'url':'https://github.com/graphql-go/graphql',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-go.md',
      'technology':['Go']
    },
    'graphql-java':{
      'name':'graphql-java - GraphQL for Java',
      'url':'https://www.graphql-java.com',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-java.md',
      'technology':['Java']
    },
    'juniper':{
      'name':'Juniper - GraphQL for Rust',
      'url':'https://graphql-rust.github.io',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/juniper.md',
      'technology':['Rust']
    },
    'sangria':{
      'name':'Sangria - GraphQL for Scala',
      'url':'https://sangria-graphql.github.io',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/sangria.md',
      'technology':['Scala']
    },
    'flutter':{
      'name':'Flutter - GraphQL for Dart',
      'url':'https://github.com/zino-app/graphql-flutter',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/gql-dart.md',
      'technology':['Dart']
    },
    'dianajl':{
      'name':'Diana.jl - GraphQL for Julia',
      'url':'https://github.com/neomatrixcode/Diana.jl',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/diana.md',
      'technology':['Julia']
    },
    'strawberry':{
      'name':'Strawberry - GraphQL for Python',
      'url':'https://github.com/strawberry-graphql/strawberry',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/strawberry.md',
      'technology':['Python']
    },
    'tartiflette':{
      'name':'tartiflette - GraphQL for Python',
      'url':'https://github.com/tartiflette/tartiflette',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/tartiflette.md',
      'technology':['Python']
    },
    'dgraph':{
      'name':'Dgraph',
      'url':'https://dgraph.io/',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/dgraph.md',
      'technology':['JavaScript']
    },
    'directus':{
      'name':'Directus',
      'url':'https://directus.io/',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/directus.md',
      'technology':['TypeScript']
    },
    'graphql_yoga':{
      'name':'GraphQL Yoga',
      'url':'https://github.com/dotansimha/graphql-yoga',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-yoga.md',
      'technology':['TypeScript']
    },
    'lighthouse':{
      'name':'Lighthouse',
      'url':'https://github.com/nuwave/lighthouse',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/lighthouse.md',
      'technology':['PHP']
    },
    'agoo':{
      'name':'Agoo',
      'url':'https://github.com/ohler55/agoo',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/agoo.md',
      'technology':['Ruby']
    },
    'mercurius':{
      'name':'mercurius',
      'url':'https://github.com/mercurius-js/mercurius',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/mercurius.md',
      'technology':['JavaScript', 'Node.js', 'TypeScript']
    },
    'morpheus-graphql':{
      'name':'morpheus-graphql',
      'url':'https://github.com/morpheusgraphql/morpheus-graphql',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/morpheus-graphql.md',
      'technology':['Haskell']
    },
    'lacinia':{
      'name':'lacinia',
      'url':'https://github.com/walmartlabs/lacinia',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/lacinia.md',
      'technology':['Clojure']
    },
    'caliban':{
      'name':'caliban',
      'url':'https://github.com/ghostdogpr/caliban',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/caliban.md',
      'technology':['Scala'],
      },
    'jaal':{
      'name':'jaal',
      'url':'https://github.com/appointy/jaal',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/jaal',
      'technology':['Golang']
    },
    'absinthe-graphql':{
      'name':'absinthe-graphql',
      'url':'https://github.com/absinthe-graphql/absinthe',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/absinthe-graphql.md',
      'technology':['Elixir']
    },
    'graphql-dotnet':{
      'name':'graphql-dotnet',
      'url':'https://github.com/graphql-dotnet/graphql-dotnet',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/graphql-dotnet.md',
      'technology':['C#', '.NET']
    },
    'pg_graphql':{
      'name':'pg_graphql',
      'url':'https://supabase.github.io/pg_graphql',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/pg_graphql.md'  ,
      'technology':['Rust']
    },
    'tailcall':{
      'name':'tailcall',
      'url':'https://tailcall.run',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/tailcall.md',
      'technology':['Rust']
    },
    'hotchocolate':{
      'name':'hotchocolate',
      'url':'https://chillicream.com/docs/hotchocolate/v13',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/hotchocolate.md',
      'technology':['C#', '.NET']
    },
    'inigo':{
      'name':'inigo',
      'url':'https://inigo.io',
      'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/inigo.md',
      'technology':['Go']
    },
    'ballerina':{
       'name':'ballerina',
       'url':'https://github.com/ballerina-platform/module-ballerina-graphql',
       'ref':'https://github.com/nicholasaleks/graphql-threat-matrix/blob/master/implementations/ballerina.md',
       'technology':['Ballerina', 'Java']
    }
  }

def user_confirmed(choice):
  if choice in ('yes', 'y'):
     return True
  return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import conf

from graphw00f.helpers import (
  get_time,
  draw_art,
  get_engines,
  user_confirmed,
  read_custom_wordlist,
  possible_graphql_paths,
  bcolors
)

from time import sleep
from urllib.parse import urlparse
from optparse import OptionParser

from version import VERSION
from graphw00f.lib import (
  GRAPHW00F,
  GraphQLDetectionFailed
)


def main():
    parser = OptionParser(usage='%prog -d -f -t http://example.com')
    parser.add_option('-r', '--noredirect', action='store_false', dest='followredirect', default=True,

                            help='Do not follow redirections given by 3xx responses')
    parser.add_option('-t', '--target', dest='url', help='target url with the path')
    parser.add_option('-f', '--fingerprint', dest='fingerprint', default=False, action='store_true', help='fingerprint mode')
    parser.add_option('-d', '--detect', dest='detect', default=False, action='store_true', help='detect mode')
    parser.add_option('-p', '--proxy', dest='proxy', default=None, help='HTTP(S) proxy URL in the form http://user:pass@host:port')
    parser.add_option('-T', '--timeout', dest='timeout', default=10, help='Request timeout in seconds')
    parser.add_option('-o', '--output-file', dest='output_file',
                            help='Output results to a file (CSV)', default=None)
    parser.add_option('-l', '--list', dest='list', action='store_true', default=False,
                            help='List all GraphQL technologies graphw00f is able to detect')
    parser.add_option('-u', '--user-agent', dest='useragent', default=None, help='Custom user-agent to use (overrides the one from headers configuration)')
    parser.add_option('-H', '--header', dest='header', action='append', default=[], help='Custom headers to send (e.g. "Authorization: Bearer ey...").')
    parser.add_option('-w', '--wordlist', dest='wordlist', default=False, help='Path to a list of custom GraphQL endpoints')
    parser.add_option('--version', '-v', dest='version', action='store_true', default=False,
                            help='Print out the current version and exit.')
    options, args = parser.parse_args()

    if options.list:
      print(draw_art())
      count = 0
      for k, v in get_engines().items():
        count += 1
        print('{index}. {name} ({technology})'.format(
                                            index=count,
                                            name=v['name'],
                                            technology=', '.join(v['technology']))
                                           )
      sys.exit(0)

    if options.version:
      print('version:', VERSION)
      sys.exit(0)

    if not options.url:
      parser.print_help()
      sys.exit(1)

    if not options.detect and not options.fingerprint:
      parser.print_help()
      sys.exit(1)

    proxies = None
    if options.proxy:
      proxies = {
          'http': options.proxy,
          'https': options.proxy
      }

    if not isinstance(options.timeout, int):
      options.timeout = 10

    headers = {}
    headers["User-Agent"] = conf.HEADERS["User-Agent"]

    # Read configuration headers
    for header in conf.HEADERS:
        headers[header] = conf.HEADERS[header]

    for header in options.header:
      key, value = header.split(": ")
      headers[key] = value    

    if options.useragent:
      headers["User-Agent"] = options.useragent

    print(headers)
    g = GRAPHW00F(follow_redirects=options.followredirect,
                  headers=headers,
                  cookies=conf.COOKIES,
                  timeout=options.timeout,
                  proxies=proxies)
    url = options.url

    url_scheme = urlparse(url).scheme
    url_netloc = urlparse(url).netloc
    wordlist = possible_graphql_paths()
    detected = False
    print(draw_art())

    if url_scheme not in ('http', 'https'):
      print('URL is missing a scheme (http|https)')
      sys.exit(1)

    if not url_netloc:
      print('url {url} does not seem right.'.format(url=url))
      sys.exit(1)

    if options.detect:
      if options.wordlist:
        wordlist = read_custom_wordlist(options.wordlist)

      for endpoint in wordlist:
        target = url + endpoint
        print('[*] Checking {}'.format(target))
        try:
          g.check(target)
          print('[!] Found GraphQL at {}'.format(target))
          url = target
          detected = True

          if not options.fingerprint:
            sys.exit(0)

          break
        except GraphQLDetectionFailed:
          continue
      if not detected:
        print('[x] Could not find GraphQL anywhere.')
        sys.exit(1)
    else:
      print('[*] Checking if GraphQL is available at {url}...'.format(url=url))  
      fingerprint = None
      try:
        if g.check(url):
          print('[!] Found GraphQL.')
      except GraphQLDetectionFailed:
          print(bcolors.FAIL + '[x] Could not determine the existence of GraphQL (Error: GraphQLDetectionFailed)' + bcolors.ENDC)
          print('[*] Continue anyway? [y/n]'.format(url=url))
          choice = input().lower()
          if not user_confirmed(choice):
            print('Quitting.')
            sys.exit(1)

    print('[*] Attempting to fingerprint...')
    result = g.execute(url)

    if result:
      name = get_engines()[result]['name']
      url = get_engines()[result]['url']
      ref = get_engines()[result]['ref']
      technologies = ', '.join(get_engines()[result]['technology'])
      fingerprint = name
      print(bcolors.OKGREEN + '[*] Discovered GraphQL Engine: ({})'.format(name))
      print('[!] Attack Surface Matrix: {}'.format(ref))
      print('[!] Technologies: {}'.format(technologies))
      print('[!] Homepage: {}'.format(url))
    else:
      print('[x] Nothing was found :-(')

    if options.output_file:
      f = open(options.output_file, 'w')
      f.write('url,detected_engine,timestamp\n')
      f.write('{},{},{}\n'.format(url_netloc, fingerprint, get_time()))
      f.close()
    
    print(bcolors.ENDC + '[*] Completed.')

if __name__ == '__main__':
    main()
    
